"use strict";

// determines if it is required to traverse an object's children.
// checks if the reference already occurs in the visited map
// if not it adds the current jsonPath to the map.
// it returns [action, node]
//
// action: 
//    0 = no visitation needed.
//    1 = object to visit. 
//    2 = array to visit.
// node: either a jsonPath to the first occurence or node to visit
function shouldVisit(obj, visited, path) {
  if (typeof obj != 'object'  
    && typeof obj != 'function' ) {
      return [0, obj];
  }
  
  if (visited.has(obj)) {
    return [0, {$ref:visited.get(obj)}];
  }

  visited.set(obj, path);
  return [1+Array.isArray(obj), obj]
}

function makeVisitor(visited, stack) {
  return function visit(obj, path, tree, key) {
    const [action, node] = shouldVisit(obj, visited, path);
 
    if (action == 0) {
      tree[key] = node;
      return;
    }

    if (key) {
      tree[key] = (action == 1) ? {} : [];
      tree = tree[key];
    }

    const nextPath = (action == 1)
      ? (k => `${path}.${k}`)
      : (k => `${path}[${k}]`)

    Object.entries(node).forEach(
      ([k, nextObj]) => stack.push(() =>
        visit(nextObj, nextPath(k) , tree, k
    )));
  }
}


// encodes cycles in objects as jsonPaths in a non-recursive manner.
const encodeCycles = obj => {
  if ((typeof obj != 'object')
    && (typeof obj != 'function')) {
    return obj;
  }

  let result = Array.isArray(obj) ? []: {};

  let visited = new Map();
  let fifo = [];
  const visit = makeVisitor(visited, fifo);
  visit(obj, '$', result);

  while (fifo.length) {
    fifo.shift()();
  }

  return result;
}


module.exports = {
  encodeCycles,
  install() {
    // For anything other than polyfills;
    // it is bad practice to extend functionality of build in types.
    // Main reason is possible name collisions if more libraries follow this practice and it may brake assumptions of others.
    // Therefore extending the JSON object is an optional feature.
    JSON.encodeCycles = encodeCycles;
  }
};
