const assert = require('assert');
require('./index').install();

describe('encodeCycles()', () => {
  it('should keep values as values', () => {
    assert.equal(JSON.encodeCycles(1), 1);
    assert.equal(JSON.encodeCycles("1"), "1");
  });

  it('should keep arrays as arrays', () => {
    assert.deepEqual(
      JSON.encodeCycles([1, 2, 3]), 
      [1, 2, 3]);
  });

  it('should iterate through entries in functions', () => {
    let f = () => 1;
    f.x=1;
    f.y=2;
    assert.deepEqual(
      JSON.encodeCycles(f), 
      {x:1, y:2}
    );
  });

  it('should detect root-cycle and replace with $', () => {
    let obj = {
      a: 1,
      b: [1, 2]
    };

    obj.b.push(obj);

    assert.deepEqual(
      JSON.encodeCycles(obj), {
        a: 1,
        b: [1, 2, {$ref:'$'}]
    });
  });

  it('should handle arrays in objects', () => {
    let o = {a:[], b:[]};
    assert.deepEqual(
      JSON.encodeCycles(o), 
      o 
    );
  });


  it('should handle objects in arrays', () => {

    let o = [ {}, {}, {} ];

    assert.deepEqual(
      JSON.encodeCycles(o), 
      o 
    );
  });

  it('should detect reference inside sibling in array and replace with a jsonPath', () => {
    let o1 = {
      a: 1,
      b: [1, 2]
    };

    let o2 = {
      a: 2,
      b: [3, 4, o1]
    };

    o2.b.push(o2);

    let o = [o1, o2];

    const expected = [{
      a: 1,
      b: [1, 2]
    }, {
      a: 2,
      b: [3, 4, 
        {$ref: '$[0]'}, 
        {$ref: '$[1]'}] 
    }];
  

    assert.deepEqual(
      JSON.encodeCycles(o), 
      expected
    );
  });

  it('should detect reference inside sibling in object and replace with a jsonPath', () => {

    let o1 = {
      a: 1,
      b: [1, 2]
    };

    let o2 = {
      a: 2,
      b: [3, 4, o1]
    };

    o2.b.push(o2);

    let o = {o1, o2};

    const expected = {
      o1:{
        a: 1,
        b: [1, 2]
      }, 
      o2: {
        a: 2,
        b: [3, 4, 
          {$ref: '$.o1'}, 
          {$ref: '$.o2'}] 
      }
    };
  

    assert.deepEqual(
      JSON.encodeCycles(o), 
      expected
    );
  });


  it('should handle deeply nested objects', () => {
    let x = {};
    let o = x;
    for (let i=0; i<1000; ++i) {
      x[0] = {};
      x = x[0];
    }

    assert.deepEqual(
      JSON.encodeCycles(o),
      o
    );
  });


})
